import time

import redis
from nose.tools import raises

from rratelimit.exceptions import UnsupportedRedisVersion

class LimiterTestBase(object):

    def __init__(self, limiter_class, **kwargs):
        self.action = 'action'
        self.limit = 10
        self.period = 0.25

        r = redis.StrictRedis(host='localhost', port=6379)
        self.redis = r
        self.limiter = limiter_class(r, self.action, self.limit, self.period,
                                     **kwargs)

    def tearDown(self):
        self.limiter.clear('test')

    def test_get_key(self):
        assert self.limiter.get_key('test') == 'rratelimit:action:test'

    def test_clear(self):
        for r in range(self.limit*2):
            self.limiter.insert('test')
        self.limiter.clear('test')
        assert not self.limiter.check('test')

    def test_unlimited(self):
        "Test that check returns False when the limit is not exceeded"
        for r in range(self.limit-1):
            self.limiter.insert('test')
        assert not self.limiter.check('test')

    def test_limited(self):
        "Test that check returns True when the limit is exceeded"
        for r in range(self.limit):
            self.limiter.insert('test')
        assert self.limiter.check('test')

    def test_expires(self):
        "Test that insertions expire after `per` time has elapsed"
        for r in range(self.limit):
            self.limiter.insert('test')
        assert self.limiter.check('test')
        time.sleep(self.period)
        assert not self.limiter.check('test')

    def test_exact(self):
        for r in range(self.limit-1):
            self.limiter.insert('test')
        # kinda racy, should be fine as long as next 4
        # operations take less than 0.01s...
        assert not self.limiter.check('test')
        time.sleep(self.period-0.1)
        self.limiter.insert('test')
        assert self.limiter.check('test')
        time.sleep(0.1)
        assert not self.limiter.check('test')

    def test_pipeline(self):
        p = self.redis.pipeline()
        for r in range(self.limit):
            self.limiter.insert('test', p)
        self.limiter.check('test', p)
        assert p.execute()[-1]

    @raises(UnsupportedRedisVersion)
    def test_invalid_version(self):
        class Redismock(object):
            def info(self):
                return {'redis_version': '2.4.0'}
        mock = Redismock()
        self.limiter.check_ver(mock)

    def test_checked_insert(self):
        for r in range(self.limit):
            assert self.limiter.checked_insert('test')
        assert not self.limiter.checked_insert('test')


from rratelimit.utils import dtime

def test_dtime():
    assert dtime(0, 10, 1) == 0
    assert dtime(0.99, 10, 1) == 0
    assert dtime(1, 10, 1) == 1
    assert dtime(11, 10, 1) == 1
    assert dtime(1, 10, .5) == 2
    assert dtime(1.49, 10, .5) == 2


import time

from .utils import LimiterTestBase
from rratelimit import SimpleLimiter

class TestSimpleLimiter(LimiterTestBase):
    def __init__(self):
        super(TestSimpleLimiter,self).__init__(SimpleLimiter)

    def test_exact(self):
        # SimpleLimiter is too inaccurate, exact will almost always fail
        pass

    def test_get_key(self):
        # SimpleLimiter's get key works differently than others
        keys = [
            'rratelimit:action:test:0',
            'rratelimit:action:test:1',
            'rratelimit:action:test:2'
        ]
        keys.remove(self.limiter.get_key('test'))
        time.sleep(self.period)
        keys.remove(self.limiter.get_key('test'))
        time.sleep(self.period)
        keys.remove(self.limiter.get_key('test'))
        assert len(keys) == 0



class UnsupportedRedisVersion(ValueError):
    def __init__(self, version):
        message = ("rratelimit requires redis-server >= 2.6.0, version {}"
                   " currently installed".format(version))
        ValueError.__init__(self, message)



from .utils import LimiterTestBase
from rratelimit import HashLimiter

class TestHashLimiter(LimiterTestBase):
    def __init__(self):
        super(TestHashLimiter,self).__init__(HashLimiter, accuracy=50)


from setuptools import setup
import multiprocessing

setup(name='rratelimit',
      version='0.0.3',
      description='Rate limiting classes for redis and redis-py',
      url='http://github.com/HeyImAlex/rratelimit',
      author='Alex Guerra',
      author_email='alex@heyimalex.com',
      classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
      ],
      license='MIT',
      packages=['rratelimit'],
      zip_safe=False,
      test_suite = 'nose.collector',
      tests_require = ['nose>=1.0', 'redis']
    )


import os

from .exceptions import UnsupportedRedisVersion

basepath = os.path.abspath(os.path.dirname(__file__))


class AbstractLimiter(object):

    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def get_key(self, actor):
        return ':'.join(['rratelimit', self.action, actor])

    def alert(self, *args, **kwargs):
        raise NotImplementedError

    def exceeded(self, *args, **kwargs):
        raise NotImplementedError

    def clear(self, *args, **kwargs):
        raise NotImplementedError


class LuaLimiter(AbstractLimiter):

    def __setattr__(self, name, value):
        if name == 'redis':
            self.check_ver(value)
            self.register_all(value)
        super(LuaLimiter, self).__setattr__(name, value)

    def register_script(self, redis, scriptname):
        """Register script located at ./lua/<scriptname>.lua"""
        path = os.path.join(basepath, 'lua', scriptname+'.lua')
        return redis.register_script(open(path).read())

    def register_all(self, *args, **kwargs):
        """Registers all lua scripts on redis instance'
           Must be overridden by child."""
        raise NotImplementedError

    def check_ver(self, redis):
        def versiontuple(v):
            return tuple(map(int, (v.split("."))))
        version = redis.info()['redis_version']
        if versiontuple(version) < versiontuple("2.6.0"):
            raise UnsupportedRedisVersion(version)


def dtime(timestamp, slots, period):
    """ Discrete time.

    Takes time and converts into into
    discrete cyclical blocks."""
    return int(timestamp / period) % slots



import time

from .utils import LuaLimiter, dtime


class ListLimiter(LuaLimiter):

    def __init__(self, redis, action, limit, period):
        self.redis = redis
        self.action = action
        self.limit = limit
        self.period = period

    def register_all(self, redis):
        self._insert = self.register_script(redis, 'list_insert')
        self._check = self.register_script(redis, 'list_check')
        self._checked_insert = self.register_script(redis,
                                                    'list_checked_insert')

    def clear(self, actor):
        self.redis.delete(self.get_key(actor))

    def insert(self, actor, client=None):
        self._insert(
            keys=[self.get_key(actor), ],
            args=[time.time(), self.period],
            client=client
        )

    def check(self, actor, client=None):
        return bool(self._check(
            keys=[self.get_key(actor), ],
            args=[time.time(), self.period, self.limit],
            client=client
        ))

    def checked_insert(self, actor, client=None):
        return bool(self._checked_insert(
            keys=[self.get_key(actor), ],
            args=[time.time(), self.period, self.limit],
            client=client,
        ))


class HashLimiter(LuaLimiter):

    def __init__(self, redis, action, limit, period, accuracy):
        self.redis = redis
        self.action = action
        self.limit = limit
        self.period = period
        self.accuracy = accuracy
        self.total_buckets = accuracy*2
        self.bucket_size = period/float(accuracy)

    def register_all(self, redis):
        self._insert = self.register_script(redis, 'hash_insert')
        self._check = self.register_script(redis, 'hash_check')
        self._checked_insert = self.register_script(redis,
                                                    'hash_checked_insert')

    def clear(self, actor):
        self.redis.delete(self.get_key(actor))

    def current_bucket(self):
        return dtime(time.time(), self.total_buckets, self.bucket_size)

    def insert(self, actor, client=None):
        self._insert(
            keys=[self.get_key(actor), ],
            args=[
                self.current_bucket(),
                self.total_buckets,
                self.period],
            client=client
        )

    def check(self, actor, client=None):
        return bool(self._check(
            keys=[self.get_key(actor), ],
            args=[self.current_bucket(), self.total_buckets, self.limit],
            client=client
        ))

    def checked_insert(self, actor, client=None):
        return bool(self._checked_insert(
            keys=[self.get_key(actor), ],
            args=[
                self.current_bucket(),
                self.total_buckets,
                self.limit,
                self.period],
            client=client
        ))


class SimpleLimiter(LuaLimiter):

    def __init__(self, redis, action, limit, period):
        self.redis = redis
        self.action = action
        self.limit = limit
        self.period = period

    def get_key(self, actor):
        orig = super(SimpleLimiter, self).get_key(actor)
        current = dtime(time.time(), 3, self.period)
        return "{}:{}".format(orig, current)

    def clear(self, actor):
        orig = super(SimpleLimiter, self).get_key(actor)
        self.redis.delete(orig + ":0")
        self.redis.delete(orig + ":1")
        self.redis.delete(orig + ":2")

    def register_all(self, redis):
        self._insert = self.register_script(redis, 'simple_insert')
        self._check = self.register_script(redis, 'simple_check')
        self._checked_insert = self.register_script(redis,
                                                    'simple_checked_insert')

    def insert(self, actor, client=None):
        self._insert(
            keys=[self.get_key(actor), ],
            args=[self.period, ],
            client=client
        )

    def check(self, actor, client=None):
        return bool(self._check(
            keys=[self.get_key(actor), ],
            args=[self.limit, ],
            client=client
        ))

    def checked_insert(self, actor, client=None):
        return bool(self._checked_insert(
            keys=[self.get_key(actor), ],
            args=[self.limit, self.period],
            client=client
        ))
